// Demonstrations of binary search
import java.util.*;
public class BinarySearch{

  // Iterative
  // Use binary search to locate integer key in array a
  public static int binarySearch(int a[], int key){
    int left=0, right=a.length-1;
    int mid = 0;
    while(left <= right){
      mid = (left+right)/2;
      if(key == a[mid]){
	return mid;
      }else if(key < a[mid]){
	right = mid-1;
      }
      else{
	left = mid+1;
      }
    }
    return -1;
  }

  // Iterative
  // Use binary search to locate integer key in array a
  public static<T extends Comparable<T>> int binarySearch(T a[], T key){
    int left=0, right=a.length-1;
    int mid = 0;
    while(left <= right){
      mid = (left+right)/2;
      //      System.out.printf("Searching %d %d %d %s\n",mid,left,right, a[mid]);
      if(key.compareTo(a[mid])==0){
	return mid;
      }else if(key.compareTo(a[mid]) < 0){
	right = mid-1;
      }
      else{
	left = mid+1;
      }
    }
    return -1;
  }

  
  // Iterative
  // Use binary search to locate integer key in array a
  public static<T> int binarySearch(T a[], T key, Comparator<T> cmp){
    int left=0, right=a.length-1;
    int mid = 0;
    while(left <= right){
      mid = (left+right)/2;
      //      System.out.printf("Searching %d %d %d %s\n",mid,left,right, a[mid]);
      if(cmp.compare(key,(a[mid]))==0){
	return mid;
      }else if(cmp.compare(key,a[mid]) < 0){
        right = mid-1;
      }
      else{
        left = mid+1;
      }
    }
    return -1;
  }


  static class MeaningOfLife implements Comparator<Integer>{
    public int compare(Integer x, Integer y){
      if(x==42 && y!=42){ return -1; }
      else if(x!=42 && y==42){ return +1; }
      else{
        return x-y;
      }
    }
  }


  // Easy-to-use recursive version which calls a helper
  public static int binarySearchR(int a[], int key){
    return binarySearchR(a,key,0,a.length-1);
  }

  // Helper method which does work of searching, repeatedly halving
  // array search area
  public static int binarySearchR(int a[], int key,
                                  int left, int right){
    if(left > right){
      return -1;
    }
    int mid = (left+right)/2;
    if(key == a[mid]){
      return mid;
    }else if(key < a[mid]){
      return
        binarySearchR(a,key,left,mid-1);
    }
    else{
      return
        binarySearchR(a,key,mid+1,right);
    }
  }
    


  public static void main(String args[]){
    Integer [] arr = {1,4,8,9,12,22,42,64,65};
    
    Comparator<Integer> mol = new MeaningOfLife();

    Arrays.sort(arr, mol);

    System.out.println(Arrays.toString(arr));

    System.out.println(java.util.Arrays.toString(arr));
    System.out.println( " 1: " + binarySearch( arr, 1, mol) );
    System.out.println( "42: " + binarySearch( arr,42, mol) );
    System.out.println( " 8: " + binarySearch( arr, 8, mol) );
    System.out.println( "10: " + binarySearch( arr,10, mol) );
    System.out.println( "22: " + binarySearch( arr,22, mol) );
    System.out.println( "75: " + binarySearch( arr,75, mol) );
    System.out.println( "55: " + binarySearch( arr,55, mol) );
    System.out.println( "-4: " + binarySearch( arr,-4, mol) );



    // String sa[] = {"aardvark", "elephant", "giraffe", "goat", "ostrich", "pig", "zebra"};
    // System.out.printf("%s %d\n","pig", binarySearch(sa, "pig"));
    // System.out.printf("%s %d\n","aardvark", binarySearch(sa, "aardvark"));
    // System.out.printf("%s %d\n","ostrich", binarySearch(sa, "ostrich"));

  }

}
